"""

From test job
 -> Parse test plan
 -> Download/Update relevant charm versions (for now all to latest)
 -> Load environment
 ->

"""

import logging
import json
import os
import shutil
import subprocess
import sys
import zookeeper

from twisted.internet import reactor


REPO_DIR = "/tmp/test-repo"


log = logging.getLogger("juju.loader")

############################


def deploy(repo_dir, charm, service=None):
    """Deploy a charm as a service.
    """
    args = ["juju", "deploy",
            "--repository", repo_dir,
            "local:%s" % charm]
    if service:
        args.append("%s" % service)
    output = subprocess.check_output(args, stderr=subprocess.STDOUT)
    log.info("Deployed %s" % (service or charm))
    return output


def destroy(service_name):
    output = subprocess.check_output(
        ["juju", "destroy-service", service_name],
        stderr=subprocess.STDOUT)
    log.info("Destroyed %s" % (service_name))
    return output


def status(with_stderr=False):
    """Get a status dictionary.
    """
    args = ["juju", "status", "--format=json"]
    if with_stderr:
        output = subprocess.check_output(
            args, stderr=subprocess.STDOUT)
        return output
    else:
        output = subprocess.check_output(args, stderr=open("/dev/null"))
        return json.loads(output)


def add_relation(relation):
    """Add a relation.
    """
    args = ["juju", "add-relation"]
    args.extend(relation[:2])
    output = subprocess.check_output(args, stderr=subprocess.STDOUT)
    log.info("Added relation %s -> %s (%s)" % (
        relation[0], relation[1], relation[2]))
    return output

############################


def setup_repository(services, repo_dir, series):
    if not os.path.exists(repo_dir):
        os.makedirs(repo_dir)

    if not os.path.isdir(repo_dir):
        raise ValueError("Invalid repository directory %s" % repo_dir)

    series_dir = os.path.join(repo_dir, series)

    if not os.path.exists(series_dir):
        os.makedirs(series_dir)

    for service in services:
        _bzr_branch(service, series_dir)


def _bzr_branch(info, series_dir):
    branch_dir = os.path.join(series_dir, info['charm'])
    log.debug("Branching charm lp:%s to %s", info["branch_spec"], branch_dir)

    if not os.path.exists(branch_dir):
        subprocess.check_output(
            ["/usr/bin/bzr", "co", "-q", "--lightweight",
             "lp:%s" % info["branch_spec"], branch_dir],
            stderr=subprocess.STDOUT)

    output = subprocess.check_output(
        ["/usr/bin/bzr", "revision-info"], cwd=branch_dir,
        stderr=subprocess.STDOUT)
    revno, revid = output.strip().split()

    if revid == info["revid"]:
        log.info(
            "Retrieved charm %(charm)s from %(branch_spec)s@%(revid)s" % info)

        return
    log.debug("Updating branch lp:%s", info["branch_spec"])
    try:
        output = subprocess.check_output(
            ["/usr/bin/bzr", "up", "-r", "revid:%s" % info["revid"]],
            stderr=subprocess.STDOUT)
    except:
        log.warning("Unable to fetch revision %s of %s ",
                    info['revid'], info['branch_spec'])

    log.info(
        "Retrieved charm %(charm)s from %(branch_spec)s@%(revid)s" % info)


def verify_bootstrapped():
    try:
        status(with_stderr=True)
    except:
        return False
    return True


def clean_juju_state(services):
    from twisted.internet.defer import inlineCallbacks
    from juju.environment.config import EnvironmentsConfig

    env_config = EnvironmentsConfig()
    env_config.load_or_write_sample()
    environment = env_config.get_default()

    @inlineCallbacks
    def _clean_juju_state():
        zookeeper.set_debug_level(0)
        provider = environment.get_machine_provider()
        storage = provider.get_file_storage()

        client = yield provider.connect()
        charms = yield client.get_children("/charms")

        # Clear out any cached charm state in zookeeper
        for s in services:  # XXX fuzzy match..
            for c in charms:
                if s in c:
                    yield client.delete("/charms/%s" % c)

        # Clear out any cached storage state
        for f in list(os.listdir(storage._path)):
            for s in services:
                if s in f:
                    os.remove(os.path.join(storage._path, f))

        reactor.stop()

    reactor.callWhenRunning(_clean_juju_state)
    reactor.run()


class EnvironmentLayer(object):

    def __init__(self, repo_dir, destroy_repo=False):
        self.repo_dir = repo_dir
        self.destroy_repo = destroy_repo

    @property
    def _state_path(self):
        return os.path.join(self.repo_dir, "test-state")

    def reset(self):
        if not os.path.exists(self._state_path):
            return
        with open(self._state_path) as fh:
            previous = json.loads(fh.read())
        current = status()

        current = current['services'].keys()
        previous = previous['services'].keys()

        added = set(current) - set(previous)
        for a in added:
            destroy(a)

        if self.destroy_repo:
            shutil.rmtree(self.repo_dir)

        # Still need to clear out the charms from zk
        clean_juju_state(added)

    def error(self, error, msg):
        if not isinstance(error, subprocess.CalledProcessError):
            log.warning(msg)
            return
        if error.cmd[:2] == ["juju", "add-relation"]:
            log.warning("Invalid plan\n\n%s" % error.output)
            return False

        if error.cmd[:2] == ["juju", "deploy"]:
            log.warning("Problem deploying %s\n\n%s" % (
                error.cmd, error.output))
            return False

        log.warning("Unknown problem %s\n\n%s" % (
            error.cmd.args, error.output))
        return False

    def __enter__(self):
        with open(os.path.join(self.repo_dir, "test-state"), "w") as fh:
            fh.write(json.dumps(status()))
        return self.repo_dir

    def __exit__(self, type, value, tb):
        capture = False
        from traceback import format_exc
        if (type, value, tb) == (None, None, None):
            pass
        else:
            error_message = format_exc()
            capture = self.error(value, error_message)
        self.reset()
        return capture


def load_test_plan(test_plan):
    if not verify_bootstrapped():
        print "Environment not bootstrapped"
        return

    setup_repository(
        test_plan['services'], REPO_DIR, test_plan['series'])

    environment = EnvironmentLayer(REPO_DIR)

    with environment as repo_dir:
        for service in test_plan["services"]:
            deploy(repo_dir, service['charm'], service['charm'])

        for relation in test_plan["relations"]:
            add_relation(relation)


def main(args):
    logging.basicConfig(level=logging.DEBUG)
    test_plan_path = args.pop()
    with open(test_plan_path) as fh:
        test_plan = json.loads(fh.read())
    load_test_plan(test_plan)


if __name__ == '__main__':
    try:
        main(sys.argv[1:])
    except:
        import pdb, traceback, sys
        traceback.print_exc()
        pdb.post_mortem(sys.exc_info()[-1])
